__all__ = [
    'Event',
    'Gladius',
]

import os
import asyncio
import inspect
from collections import defaultdict
from typing import TypedDict, Callable

try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass

from aiohttp import web
from aiohttp.web import middleware
from aiohttp.web_urldispatcher import SystemRoute

from .consts import EVENT_HANDLER_EVENT_TYPE_MAP

class Event(TypedDict, total=False):
    sf_id: str
    # type: str
    # ...

class Gladius:
    app: web.Application
    callbacks: dict[str, dict[str, tuple['Component', Callable]]] # {sf_id: {event_type: [component, func]}}

    def __init__(self):
        @middleware
        async def gladius_middleware(request: web.Request, handler: Callable) -> web.Response:
            # print(request, handler)
            req: dict

            if request.method == 'GET':
                req = dict(request.query)
            else:
                req = await request.json()

            response: web.Response

            # TODO: check line below for better request handling
            # if inspect.ismethod(handler) and hasattr(handler, '__self__') and isinstance(handler.__self__, SystemRoute):
            if inspect.ismethod(handler):
                response = await handler(request)
                return response

            res: str | dict = await handler(self, req)

            if isinstance(res, str):
                response = web.Response(text=res, content_type='text/html')
            elif isinstance(res, dict):
                response = web.json_reponse(res)
            else:
                raise ValueError(res)

            return response

        self.app = web.Application(middlewares=[gladius_middleware])
        
        self.app.add_routes([
            web.static('/static/gladius', os.path.join(os.path.split(__loader__.path)[0], 'static')),
            web.get('/api/1.0/_event/{event_type}/{sf_id}', self.get_api_1_0__event),
            web.post('/api/1.0/_event/{event_type}/{sf_id}', self.post_api_1_0__event),
        ])

        if os.path.exists('static'):
            self.app.add_routes([
                web.static('/static', 'static'),
            ])

        self.callbacks = defaultdict(dict)

    def add_route(self, method, path, handler):
        routes = [web.route(method.lower(), path, handler)]
        self.app.add_routes(routes)

    def route(self, path, component: 'Component'):
        async def _handler(sf, req) -> str:
            rendered_component: str = component.render()
            return rendered_component

        self.app.router.add_get(path, _handler)

    async def get_api_1_0__event(self, request: web.Request) -> web.Response:
        event_type: str = request.match_info['event_type']
        sf_id: str = request.match_info['sf_id']
        # print('get_api_1_0__event:', event_type, sf_id)
        component, callback = self.callbacks[sf_id][event_type]
        await callback(component, {})
        res: str = component.render()
        # print('get_api_1_0__event:', res)
        return web.Response(text=res, content_type='text/html')

    async def post_api_1_0__event(self, request: web.Request) -> web.Response:
        event_type: str = request.match_info['event_type']
        sf_id: str = request.match_info['sf_id']
        event: Event = request.headers['Triggering-Event']
        # print('post_api_1_0__event:', event_type, sf_id, event)
        component, callback = self.callbacks[sf_id][event_type]
        res: dict = await callback(component, event)
        # print('post_api_1_0__event:', res)
        return web.json_response(res)

    def get_app(self):
        return self.app

    def run_app(self, *args, **kwargs):
        web.run_app(self.app, *args, **kwargs)



__all__ = [
    'Page',
    'Navbar',
    'NavbarButton',
    'Link',
    'Flex',
    'VFlex',
    'Card',
    'Join',
    'VJoin',
    'Button',
    'Table',
    'Text',
    'Artboard',
    'Divider',
    'Footer',
    'FooterTitle',
    'Hero',
    'HeroContent',
    'Indicator',
    'IndicatorItem',
    'Stack',
    'Toast',
    'Alert',
    'DaisyUI',
]

from typing import Self

from .gladius import Gladius, Event
from .component import Component, ComponentLibrary
from . import html5

class Page(Component):
    default_class: str = 'container mx-auto'
    title: str

    def __init__(self, *args, title: str='Gladius', favicon: str='/static/gladius/favicon.png', **kwargs):
        super().__init__(*args, **kwargs)
        self.title = title
        self.favicon = favicon

    def render(self) -> str:
        rendered_children = '\n'.join(c.render() for c in self.children)

        return f'''
            <!doctype html>
            <html lang="en-US">
                <head>
                    <meta charset="utf-8" />
                    <meta name="viewport" content="width=device-width" />
                    <link rel="shortcut icon" type="image/png" href="{self.favicon}" />
                    <title>{self.title}</title>

                    <!-- daisyui -->
                    <link href="https://cdn.jsdelivr.net/npm/daisyui@3.7.7/dist/full.css" rel="stylesheet" type="text/css" />

                    <!-- tailwind -->
                    <script src="https://cdn.tailwindcss.com?plugins=forms,typography,aspect-ratio"></script>

                    <!-- htmx -->
                    <script src="https://unpkg.com/htmx.org@1.9.5"></script>
                    <script src="https://unpkg.com/htmx.org/dist/ext/debug.js"></script>
                    <script src="https://unpkg.com/htmx.org/dist/ext/json-enc.js"></script>
                    <script src="https://unpkg.com/htmx.org/dist/ext/event-header.js"></script>
                    <script src="/static/gladius/multi-path-deps.js"></script>
                </head>
                <body hx-ext='multi-path-deps'>
                    <div {self.render_attrs()}>
                        {rendered_children}
                    </div>
                </body>
            </html>
        '''

class Navbar(Component):
    default_class: str = 'navbar bg-base-100'

    def render(self) -> str:
        rendered_children = '\n'.join(c.render() for c in self.children)

        return f'''
            <div {self.render_attrs()}>
                {rendered_children}
            </div>
        '''

class NavbarButton(Component):
    # NOTE: extended component
    default_class: str = 'btn btn-ghost normal-case text-xl'

    def render(self) -> str:
        rendered_children = '\n'.join(c.render() for c in self.children)

        return f'''
            <a {self.render_attrs()}>
                {rendered_children}
            </a>
        '''

class Link(html5.A):
    default_class: str = 'link'

class Flex(html5.Div):
    default_class: str = 'flex w-full gap-1'

class VFlex(Flex):
    # NOTE: extended component
    default_class: str = 'flex flex-col w-full gap-1'

class Card(html5.Div):
    default_class: str = 'card grid h-20 flex-grow bg-base-300 rounded-box place-items-center'

class Join(html5.Div):
    default_class: str = 'join'

    def add(self, child: Component) -> Self:
        if not child.has_class('join-item'):
            child.add_class('join-item')


        self.children.append(child)
        return self

class VJoin(Join):
    # NOTE: extended component
    default_class: str = 'join join-vertical'

class Button(html5.Button):
    default_class: str = 'btn'

class Table(Component):
    header: list[str]
    rows: list[list]

    def __init__(self, component_library: ComponentLibrary, header: list[str]=[], rows: list[list]=[], **kwargs):
        super().__init__(component_library, **kwargs)
        self.header = header
        self.rows = rows

        async def _ontablechange(table: Table, event: Event):
            # print('_ontablechange', table, event)
            pass

        event_type: str = '_ontablechange'
        self.attrs[event_type] = _ontablechange

    def render(self) -> str:
        rendered_header = '\n'.join([
            '<tr>',
            '\n'.join(f'<th>{h}</th>' for h in self.header),
            '</tr>',
        ])
        
        rendered_rows = '\n'.join([
            '\n'.join([
                '<tr>',
                '\n'.join(f'<td>{v}</td>' for v in row),
                '</tr>',
            ]) for row in self.rows
        ])

        return f'''
            <table class="table" {self.render_attrs()}>
                <thead class="bg-base-200">
                    {rendered_header}
                </thead>
                <tbody> 
                    {rendered_rows}
                </tbody>
            </table>
        '''

# FIXME:
class Text(Component):
    content: str

    def __init__(self, component_library: 'ComponentLibrary', content: str='', **kwargs):
        super().__init__(component_library, **kwargs)
        self.content = content

        async def _ontextchange(text: Text, event: Event):
            # print('_ontextchange', text, event)
            pass

        event_type: str = '_ontextchange'
        self.attrs[event_type] = _ontextchange

    def render(self) -> str:
        return f'''
            <span {self.render_attrs()}>
                {self.content}
            </span>
        '''

#
# Layout
#
class Artboard(html5.Div):
    default_class: str = 'artboard'

class Divider(html5.Div):
    default_class: str = 'divider'

# TODO: Drawer

class Footer(html5.Footer):
    default_class: str = 'footer'

class FooterTitle(html5.Header):
    # NOTE: extended component
    default_class: str = 'footer-title'

class Hero(html5.Div):
    default_class: str = 'hero bg-base-200'

class HeroContent(html5.Div):
    default_class: str = 'hero-content'

class Indicator(html5.Div):
    default_class: str = 'indicator'

class IndicatorItem(html5.Span):
    default_class: str = 'indicator-item'

# TODO: Mask

class Stack(html5.Div):
    default_class: str = 'stack'

class Toast(html5.Div):
    default_class: str = 'toast'

class Alert(html5.Div):
    default_class: str = 'alert'

#
# DaisyUI Component Library
#
class DaisyUI(ComponentLibrary):
    def __init__(self, ctx: Gladius):
        super().__init__(ctx)

        self.component_map: dict[str, Component] = {
            k: v
            for k, v in dict(globals()).items()
            if isinstance(v, type) and issubclass(v, Component)
        }

    def __getattr__(self, attr):
        ComponentType: type = self.get_component_type(attr)

        def _component(*args, **kwargs) -> Component:
            return ComponentType(self, *args, **kwargs)

        return _component
